library(ncdf4)
library(raster)
library(dplyr)
library(sf)
library(mapview)
library(terra)

# PARAMETERS ----
year <- 2024
home_ada <- "/home/abarrant" #/home/abarrant
input_stats_obs <- paste0(home_ada,"/esarchive/scratch/ccarnere/RScripts/eval_gene/", year, "/statistics_at_stations.RDS")
input_stats_file <- paste0(home_ada,"/esarchive/scratch/ccarnere/diagnosi/", year, "/statistics_", year, ".nc")
output_shp <- paste0(home_ada, "/esarchive/scratch/abarrant/test_netcdf/statistics_with_obs_", year, ".shp")
output_nc <- paste0(home_ada, "/esarchive/scratch/abarrant/test_netcdf/statistics_with_obs_", year, ".nc")
lambert_crs <- "+proj=lcc +lat_1=43 +lat_2=37 +lat_0=40 +lon_0=-3 +x_0=0 +y_0=0 +datum=WGS84 +units=m +no_defs"

#output directory
dir.create(dirname(output_shp), showWarnings = FALSE, recursive = TRUE)


# REPLACING OBS IN RASTER ------------------------------------------------------

#load data with terra library
raster_stack <- rast(input_stats_file)
raster_vals <- names(raster_stack)            #get variable names that match between raster and observations

#obs
stats_obs <- readRDS(input_stats_obs)
obs_vals <- names(stats_obs)
common_vals <- intersect(raster_vals, obs_vals[!obs_vals %in% c("lon", "lat")])
cat(paste0("Common values: ", paste(common_vals, collapse = ", ")), "\n\n")

#obs as spatial data
obs_sf <- st_as_sf(stats_obs, coords = c("lon", "lat"), crs = 4326)

#prepare coordinate handling (once for all variables)
obs_coords <- data.frame(
  lon = stats_obs$lon,
  lat = stats_obs$lat
)
valid_coords <- complete.cases(obs_coords)
obs_coords <- obs_coords[valid_coords, ]
obs_points <- SpatialPoints(obs_coords, proj4string = CRS("+proj=longlat +datum=WGS84"))
obs_points_transformed <- spTransform(obs_points, CRS(lambert_crs))

#create netcdf -----------------------------------------------------------------

var_names <- setdiff(names(raster_stack), c("lat", "lon", "Lambert_conformal", "crs"))
template <- rast(input_stats_file, subds = var_names[1])
crs(template) <- crs(raster_stack)
crs(template) <- lambert_crs

# Coordinate values
xvals <- xFromCol(template, 1:ncol(template))            # left to right
yvals <- yFromRow(template, 1:nrow(template))            # top to bottom (terra default)

# NetCDF dimensions
x_dim <- ncdim_def("x", "m", xvals)
y_dim <- ncdim_def("y", "m", yvals)

# Define CF-compliant CRS variable (Lambert Conformal Conic)
crs_var <- ncvar_def("crs", "", NULL, prec = "char")

# Read original NetCDF to extract units
nc_orig <- nc_open(input_stats_file)

var_defs <- list()
for (var_name in var_names) {
  # Extract units from original NetCDF
  units <- "unknown"  # default
  if (var_name %in% names(nc_orig$var)) {
    if ("units" %in% names(nc_orig$var[[var_name]])) {
      units <- ncatt_get(nc_orig, var_name, "units")$value
    }
  }
  
  var_def <- ncvar_def(var_name, units, list(x_dim, y_dim), missval = NA, prec = "float")
  var_defs[[var_name]] <- var_def
}

# Close original NetCDF
nc_close(nc_orig)

# Create NetCDF with all variables + CRS
all_vars <- c(var_defs, list(crs = crs_var))
nc_out <- nc_create(output_nc, all_vars)

# Write CF-compliant CRS metadata
ncatt_put(nc_out, "crs", "grid_mapping_name", "lambert_conformal_conic")
ncatt_put(nc_out, "crs", "standard_parallel", c(37.0, 43.0))
ncatt_put(nc_out, "crs", "latitude_of_projection_origin", 40.0)
ncatt_put(nc_out, "crs", "longitude_of_central_meridian", -3.0)
ncatt_put(nc_out, "crs", "false_easting", 0.0)
ncatt_put(nc_out, "crs", "false_northing", 0.0)
ncatt_put(nc_out, "crs", "semi_major_axis", 6378137.0)
ncatt_put(nc_out, "crs", "inverse_flattening", 298.257223563)

# Define function to properly orient matrix for NetCDF
# Terra rasters are stored top-to-bottom, NetCDF expects bottom-to-top
# So we need to flip vertically (not transpose)
orient_for_netcdf <- function(raster_layer) {
  mat <- as.matrix(raster_layer, wide = TRUE)
  # Flip vertically to match NetCDF convention (bottom-to-top)
  t(mat[nrow(mat):1, ])
}


#lists to store results -----------------------------------------------
modified_rasters <- list()
processed_obs_data <- list()

#process each common variable
for (var_name in common_vals) {
  #var_name <- common_vals[13]
  cat("Processing variable:", var_name, "\n")
  
  #extract raster layer
  terra_layer <- raster_stack[[var_name]]
  raster_layer <- raster(terra_layer)
  
  #create modified copy
  raster_modified <- raster_layer
  
  #extract cell numbers for observation points
  cell_numbers <- cellFromXY(raster_modified, coordinates(obs_points_transformed))
  valid_cells <- !is.na(cell_numbers) & valid_coords
  cell_numbers <- cell_numbers[!is.na(cell_numbers)]
  
  #get replacement values from observations
  replacement_values <- stats_obs[[var_name]][valid_coords][!is.na(cell_numbers)]
  replacement_values <- as.numeric(replacement_values)
  
  #only replace valid (non NA) values
  valid_replacements <- !is.na(replacement_values) & !is.nan(replacement_values)
  cells_to_replace <- cell_numbers[valid_replacements]
  values_to_use <- replacement_values[valid_replacements]
  
  #replace raster cells with observation values
  raster_modified[cells_to_replace] <- values_to_use
  
  #store results
  modified_rasters[[var_name]] <- raster_modified
  cat("  - Replaced", length(values_to_use), "cells\n")
  
  ncvar_put(nc_out, var_name, orient_for_netcdf(raster_modified)) ##
  ncatt_put(nc_out, var_name, "grid_mapping", "crs")
}

nc_close(nc_out)

# CONVERT NETCDF TO SHAPEFILE --------------------------------------------------

cat("Converting NetCDF to Shapefile...\n")

nc_raster <- rast(output_nc)
crs(nc_raster) <- lambert_crs


raster_points <- as.points(nc_raster)
sf_points <- st_as_sf(raster_points)
st_crs(sf_points) <- lambert_crs

#write to shapefile
st_write(sf_points, output_shp, overwrite = TRUE)




